{$ObjExportAll On}
{$A+,B-,C+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,W-,X+,Y-,Z1}
{$D-}   { Make it D+ for debug information }
{-$ObjExportAll On} { remove "-" to enable for C++ Builder 3.0!!! }
{===================================================================}
{ TSegLHA.PAS V4.00           Dual 16/32 bit version                }
{                                                                   }
{*******************************************************************}
{ Copyright (c) 1997-8, Ken McClain                                 }
{   ALL RIGHTS RESERVED                                             }
{                                                                   }
{ This source code is provided for your inspection and adaptation.  }
{ While you are free to on-sell executable applications which make  }
{ use of the source (in whole or in part) you may NOT:              }
{ a) Use this source or any part of it to make components           }
{    (DCUs, DLLs, OBJs, OCXs etc) for resale                        }
{ b) On-sell this source code or any part of it as part of any      }
{     product                                                       }
{                                                                   }
{ If you do not accept these restrictions, please contact           }
{ South Pacific Information Services Ltd for a full and immediate   }
{ refund, or to obtain a redistribution license tailored to suit    }
{ your needs. Email: software@spis.co.nz  Web: www.spis.co.nz       }
{*******************************************************************}
{                                                                   }
{ Desc: This component is derived from the TCompLHA component and   }
{       provides archive segmenting. The intent of this component   }
{       is to provide an easy method for adding backup/restore      }
{       capabilities to any Delphi program.                         }
{                                                                   }
{ History:   Version 2.0 create 5/28/97                             }
{                                                                   }
{ Changes:                                                          }
{          07/15/97 Bug Fix.  FSegPath was not being updated after  }
{                   the FSegDir value was changed.                  }
{                                                                   }
{          09/28/97 Bug Fix.  Changed global variables, BPtr and    }
{                   MyOwner, to be private variable within the
{                   TSegLHA object. This allows multiple instances  }
{                   of TSegLha to exist in the same program.        }
{                                                                   )
{          09/29/97 Version 2.50.  Changed default setting for the  }
{                   TempArchivepath to be the current directory if  }
{                   not specified.                                  }
{                                                                   }
{          11/20/97 Bug Fix.  Corrected problem with placement of   }
{                   segment files. Slash was missing between the    }
{                   FSegDir and the Segment name.                   }
{                                                                   }
{           1/22/98 Version 2.50b.  Added new property 'SegExt' to  }
{                   allow more control over the segment extension.  }
{                   Use 4GB diskfree from TCompLHA                  }
{                                                                   }
{===================================================================}

unit SegLHA;

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinTypes, WinProcs,
{$ENDIF}
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   CompLHA;

type
{===================================================================}
{ TSegLHA specific exceptions definations                           }
{===================================================================}
ESegLHAError = Class(Exception);

{$IFDEF WIN32}
{$ALIGN OFF}
{$ENDIF}

{=TSegIDHeader======================================================}
{ TSegLHA Segment ID Header record defination. One per segment.     }
{===================================================================}
  TSegIDHeader = record
    HdrID:         array[0..3] of Char;  { Always SSID }
    SegNumber:     Word;
    end;

{=TSegSetHeader=====================================================}
{ TSegLHA Segment Set Header record defination. One per segment set }
{===================================================================}
  TSegSetHeader = record
    HdrID:         array[0..3] of Char;  { Always SSET }
    CompMode:      TCompLHAMethod;       { Compress method }
    SegDate:       TDateTime;            { Segment Datetime }
    SegSize:       Longint;              { Segment Size }
    end;

{=TSegScanHeader====================================================}
{ TSegLHA Segment Scan Header record defination. One per segment set}
{===================================================================}
  TSegScanHeader = record
    HdrID:         array[0..3] of Char;  { Always SSCN }
    ScanLen:       Longint;              { Length of Scan File}
    end;

{=TSegArcHeader=====================================================}
{ TSegLHA Segment Archive Header record defination. One per         }
{ compressed archive.                                               }
{===================================================================}
  TSegArcHeader = record
    HdrID:         array[0..3] of Char;  { Always SARC }
    ArcLen:        Longint;              { Length of Archive file }
    end;

{=TFileInfo=========================================================}
{ Redefination of the TCompLHAFileInfo object for writing to the    }
{ segment file set.  * Probably a kludge *                          }
{ **NOTE** If the TCompLHAFileInfo object changes, this  code will  }
{ need to change as well.                                           }
{===================================================================}
  TFileInfo = record
    CompressedMode:   TCompLHAMethod;
    CompressedSize:   Longint;
    FullSize:         Longint;
    Datetime:         Longint;
    Attributes:       Smallint;
    HeaderLevel:      Byte;
    CRC:              Word;
    Position:         Longint;
    end;

{$IFDEF WIN32}
{$ALIGN ON}
{$ENDIF}

{===================================================================}
{ Segment Message type                                              }
{===================================================================}
  TSegMessageType = ( smTitle, smWrongSegMounted, smPrompt, smFreeSpaceMsg,
                      smSegAbortMsg, smSegOpenMsg, smScanOpenMsg,
                      smArcOpenMsg, smSegNotFoundMsg, smCorrectCancelMsg,
                      smSegIDHeaderMsg, smSegSetHeaderMsg, smSegScanHeaderMsg,
                      smSegArcHeaderMsg );

{===================================================================}
{ TSegLHA event definations                                         }
{===================================================================}
TSegChangeEvent = procedure( const SegPath: String ) of Object;
TSegExpandEvent = procedure( const CurTargetPath: String; var NewTargetPath: String;
                             var ListOfFiles: TStrings ) of Object;
TSegPromptEvent = procedure( var SegDir: String; const SegName: String;
                                 SegNumber: Word ) of Object;

{=TSegLHA===========================================================}
{ Component defination.  Derived from TCompLHA                      }
{===================================================================}
  TSegLHA = class(TCompLHA)
  private
    {properties}
    FSegBaseName:             String;    {Segment base name w/o extention}
    FSegDir:                  String;    {Segment directory path, ie, A:\}
    {New 2.50b 1/22/98}
    FSegExt:                  String;    {Segment Extension}

    FSegMessages:             TStrings;  {Message texts}
    FSegSize:                 Longint;   {Segment maximum size in bytes  }

    FSegName:                 String;    {Current segment name.          }
    FSegNumber:               Word;      {Current segment number.        }
    FSegPath:                 String;    {Current segment path.          }

    FTmpFile:                 String;    {Temporary file}
    FScanFile:                String;    {Scan file}
    FArcFile:                 String;    {Archive file}
    FSegBytes:                Longint;   {Max Segment Bytes}
    FFreeBytes:               Longint;   {Free bytes in segment}
    FTotalBytes:              Longint;   {Total bytes in segment set}
    FBytesProcessed:          Longint;   {Total bytes processed}
    FSegShowProgress:         Longint;   {Show progress percent}

    FBPtr:                    PChar;      {Bug Fix 9/28/97}
    FMyOwner:                 TComponent; {Bug Fix 9/28/97}

    {Events}
    FOnSegChange:             TSegChangeEvent;
    FOnSegExpand:             TSegExpandEvent;
    FOnSegPrompt:             TSegPromptEvent;

    {Event handler save areas}
    FSaveOnCheckFile:         TCompLHAFileNameCheckEvent;
    FSaveOnAfterProcessed:    TCompLHAAfterProcessedEvent;
    FSaveOnShowProgress:      TCompLHAShowProgressEvent;

    {Methods}
    procedure  SetSegBaseName( const SegBaseName: String );
    procedure  SetSegDir( const SegDir: String );
    {New 2.50b 1/22/98}
    procedure  SetSegExt( const SegExt: String );

    procedure  SetSegMessages( Messages: TStrings );
    procedure  SetSegSize( SegSize: Longint );

    function   MakeTmpFile( const Ext: String ): String;
    procedure  OnCheckFileStub( const originalFilePath: String;
                                var newFilePath: String;dateTime: Longint;
                                mode: TCompLHAProcessMode );
    procedure  OnAfterProcessedStub( const originalFilePath, newFilePath: String;
                                     dateTime: Longint;
                                     mode: TCompLHAProcessMode );
    procedure  OnShowProgressStub( var PercentageDone: Longint );
    procedure  DisableEventHandlers;
    procedure  EnableEventHandlers;
    function   GetSegMessage( MsgCnt: TSegMessageType ): String;
    function   GetFileSize( const FName: String ): Longint;
    procedure  SetSegmentPath;
    function   AppendSlash( const Path: String ): String;
    procedure  RemoveFile( var Path: String );
    procedure  RemoveSegFiles;
    function   CheckSpace( const Path: String ): Single; { was Longint v2.50b }
    {Changed 2.50b 1/22/98}
    function   PathPrompt( var DefDir: String; const SegNum: Integer;
                           const SegName: String ): Boolean;

    procedure  NewSegmentEvent;

    procedure  SegmentArchive;
    procedure  AssembleArchive( VerifyOnly: Boolean);
    procedure  ScanSegment;

    procedure  ReadSegSetHeader( var fh: TFileStream; var SetHdr: TSegSetHeader );
    procedure  CreateScanFile;
    procedure  ReadScanFile( var fh: TFileStream );
    procedure  WriteScanFile( var fh: TFileStream );
    procedure  ReadArchiveFile( var fh: TFileStream );
    procedure  WriteArchiveFile( var fh: TFileStream );
    function   CreateSegment: TFileStream;
    function   OpenSegment: TFileStream;
    procedure  CloseSegment( var fh: TFileStream );
    procedure  WriteToSegment( var fh: TFileStream; var Buffer; Count: Longint );
    procedure  ReadFromSegment( var fh: TFileStream; var Buffer; Count: Longint );

  protected
    { Protected declarations }

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    function    Scan: Boolean; override;
    procedure   Compress; override;
    procedure   Expand; override;
    procedure   Verify; override;

    {Runtime ReadOnly}
    property SegName:            String read FSegName;
    property SegNumber:          Word read FSegNumber;
    property SegPath:            String read FSegPath;

  published
    property SegBaseName:        String read FSegBaseName write SetSegBaseName;
    property SegDir:             String read FSegDir write SetSegDir;
    {New 2.50b 1/22/98}
    property SegExt:             String read FSegExt write SetSegExt;

    property SegMessages:        TStrings read FSegMessages write SetSegMessages;
    property SegSize:            Longint read FSegSize write SetSegSize;

    property OnSegChange:        TSegChangeEvent  read FOnSegChange
                                                  write FOnSegChange;
    property OnSegExpand:        TSegExpandEvent  read FOnSegExpand
                                                  write FOnSegExpand;
    property OnSegPrompt:        TSegPromptEvent  read FOnSegPrompt
                                                  write FOnSegPrompt;
  end;

{=Const========================================================}
{==============================================================}
const
  MINFREE     = 2048;                   {Minimum free space on segment}
  BUFSIZE     = 8192;                   {Buffer size}
  SSID        = 'SSID';                 {TSegIDHeader Tag}
  SSET        = 'SSET';                 {TSegSetHeader Tag}
  SSCN        = 'SSCN';                 {TSegScanHeader Tag}
  SARC        = 'SARC';                 {TSegArcHeader Tag}
  CR          = #10#13;
  NUL         = #0;

{==============================================================}
{ Default Segment Messages. User can change them in SegLHA     }
{==============================================================}
  SegMessageDefaults:  array[TSegMessageType] of String =
                       ('Mount Segment # ',
                        'Incorrect segment mounted!',
                        'Segment directory:',
                        'Not enough Free space available',
                        'Operation cancelled',
                        'Unable to open selected segment.',
                        'Unable to open scan file.',
                        'Unable to open archive file.',
                        'Unable to locate selected segment file.',
                        'Correct and press OK or Cancel to Abort.',
                        'Segment ID Header invalid',
                        'Segment Set Header invalid',
                        'Segment Scan Header invalid',
                        'Segment Archive Header invalid');


procedure Register;

implementation

{=Register=====================================================}
{==============================================================}
procedure Register;
begin
  RegisterComponents('Compress', [TSegLHA]);
end;

{=SetSegBaseName===============================================}
{==============================================================}
procedure TSegLHA.SetSegBaseName( const SegBaseName: String );
begin
     if (SegBaseName = '') then
       FSegBaseName := 'SegData'
     else
       FSegBaseName := SegBaseName;
end;

{=SetSegDir====================================================}
{==============================================================}
procedure TSegLHA.SetSegDir( const SegDir: String );
begin
     if (SegDir = '') then
       FSegDir := 'a:\'
     else
       FSegDir := SegDir;
end;

{New 2.50b 1/22/98}
{=SetSegExt====================================================}
{==============================================================}
procedure TSegLHA.SetSegExt( const SegExt: String );
begin
     if (Length(SegExt) > 3) then
       FSegExt := Copy(SegExt,1,3)
     else
       FSegExt := SegExt;
end;

{=SetSegMessages===============================================}
{==============================================================}
procedure TSegLHA.SetSegMessages( Messages: TStrings );
begin
     FSegMessages.Assign(Messages);
end;

{=SetSegSize===================================================}
{==============================================================}
procedure TSegLHA.SetSegSize( SegSize: Longint );
begin
     if (SegSize < 8192) then
       FSegSize := 8192
     else
       FSegSize := SegSize;
end;

{=MakeTmpFile==================================================}
{==============================================================}
function  TSegLHA.MakeTmpFile( const Ext: String ): String;
var
   CDir:    String;                               {New in 2.50}
begin
     if (TempArchivePath = '') then
        begin
        GetDir(0,CDir);                           {New in 2.50}
        TempArchivePath := CDir;                  {New in 2.50}
        end;
     Result := TempArchivePath + FSegBaseName + Ext;
     if (FileExists(Result)) then
       DeleteFile(Result);
end;

{=OnCheckFileStub==============================================}
{==============================================================}
procedure TSegLHA.OnCheckFileStub( const originalFilePath: String;
                                       var newFilePath: String;dateTime: Longint;
                                       mode: TCompLHAProcessMode );
begin
     {This is a no-op}
end;

{=OnAfterProcessedStub=========================================}
{==============================================================}
procedure  TSegLHA.OnAfterProcessedStub( const originalFilePath, newFilePath: String;
                                             dateTime: Longint;
                                             mode: TCompLHAProcessMode );
begin
     {This is a no-op}
end;

{=OnShowProgressStub===========================================}
{==============================================================}
procedure TSegLHA.OnShowProgressStub( var PercentageDone: Longint );
begin
     {This is a no-op}
end;

{=DisableEventHandlers=========================================}
{==============================================================}
procedure TSegLHA.DisableEventHandlers;
begin;
      FSaveOnCheckFile := OnCheckFile;
      FSaveOnAfterProcessed := OnAfterProcessed;
      FSaveOnShowProgress := OnShowProgress;
      OnCheckFile := OnCheckFileStub;
      OnAfterProcessed := OnAfterProcessedStub;
      OnShowProgress := OnShowProgressStub;
end;

{=EnableEventHandlers==========================================}
{==============================================================}
procedure TSegLHA.EnableEventHandlers;
begin;
      OnCheckFile := FSaveOnCheckFile;
      OnAfterProcessed := FSaveOnAfterProcessed;
      OnShowProgress := FSaveOnShowProgress;
end;

{=GetSegMessage================================================}
{==============================================================}
function TSegLHA.GetSegMessage( MsgCnt: TSegMessageType ): String;
begin
     Result := '';
     if (Ord(MsgCnt) > pred(SegMessages.Count)) then
       exit;
     Result := SegMessages[Ord(MsgCnt)];
end;

{=GetFileSize==================================================}
{==============================================================}
function TSegLHA.GetFileSize( const FName: String ): Longint;
var
  fh:        TFileStream;
begin
     Result := 0;
     if ((FName = '') or (FileExists(FName) = False)) then
       exit;
     fh := TFileStream.Create(FName,fmOpenRead);
     Result := fh.Size;
     fh.Free;
end;

{=AppendSlash==================================================}
{==============================================================}
function  TSegLHA.AppendSlash( const Path: String ): String;
var
   TmpPath:     String;
begin
     if (Path[Length(Path)] = '\') then
       TmpPath := Path
     else
       TmpPath := Path + '\';
     Result := TmpPath;
end;

{=SetSegmentPath===============================================}
{==============================================================}
procedure TSegLHA.SetSegmentPath;
var
   i:      Integer;
   Num:    String;
begin
     {Changed 2.50b 1/22/98}
     FSegName := FSegBaseName + '.';
     if (FSegExt = '') then
       begin
       Num := IntToStr(FSegNumber);
       for i := 1 to (3-Length(Num)) do
         FSegName := FSegName + '0';
       FSegName := FSegName + Num;
       end
     else
       FSegName := FSegName + FSegExt;

     FSegPath := AppendSlash(FSegDir) + FSegName;
end;

{=RemoveSegFile================================================}
{==============================================================}
procedure TSegLHA.RemoveFile( var Path: String );
begin
     if (Path <> '') then
       begin
       if (FileExists(Path) = True) then
         DeleteFile(Path);
       Path := '';
       end;
end;

{=RemoveSegFile================================================}
{==============================================================}
procedure TSegLHA.RemoveSegFiles;
begin
     RemoveFile(FTmpFile);
     RemoveFile(FScanFile);
     RemoveFile(FArcFile);
end;

{=CheckSpace===================================================}
{==============================================================}
{ amended to Single return v2.50b -- now using diskfree from TCompLHA }
function  TSegLHA.CheckSpace( const Path: String ): Single;
var
  DiskID:     Byte;
begin
     Result := 0;
     if ((Length(Path) > 1) and (Path[2] = ':')) then
       begin
       DiskID := Byte(Ord(Path[1]) - Ord('A') + 1);
       Result := DiskFree(DiskID);
       end;
end;

{Changed 2.50b 1/22/98}
{=PathPrompt===================================================}
{==============================================================}
function  TSegLHA.PathPrompt( var DefDir: String; const SegNum: Integer;
                              const SegName: String ): Boolean;
var
  SaveCursor:      TCursor;
  NewDefDir:       String;
  Title:           String;
begin
     Result := False;
     SaveCursor := Screen.Cursor;
     Screen.Cursor := crDefault;
     {Change 2.50b 1/22/98}
     Title := GetSegMessage(smTitle) + IntToStr(SegNum);

     while (True) do
       begin
       NewDefDir := DefDir;
       if (InputQuery(Title,GetSegMessage(smPrompt),newDefDir) = False) then
         begin
         Result := True;
         Break;
         end;
       if (NewDefDir <> '') then
         begin
         DefDir := NewDefDir;
         break;
         end;
       end;
     Screen.Cursor := SaveCursor;
end;

{=NewSegmentEvent==============================================}
{==============================================================}
procedure TSegLHA.NewSegmentEvent;
begin
     {=========================================================}
     { If the user supplies an OnSegPrompt event handler, call }
     { it now. Otherwise use the default event handler. Then   }
     { check to see if they aborted the operation              }
     {=========================================================}
     AbortOperation := False;
     if (Assigned(FOnSegPrompt)) then
       OnSegPrompt(FSegDir,FSegName,FSegNumber)
     else
       begin
       {Change 2.50b 1/22/98}
       AbortOperation := PathPrompt(FSegDir,FSegNumber,FSegName);
       (FMyOwner as TForm).RePaint;           {Bug Fix 9/28/97}
       end;

     if (AbortOperation = True) then
       begin
       raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );
       exit;
       end;

     {=========================================================}
     { 7/15/97 Bug Fix.                                        }
     { This corrects a problem in the initial release where    }
     { the FSegPath was not modified if the FSegDir changed    }
     { 11/20/97 Bug Fix.                                       }
     { Corrects problem where '\' was missing between FSegDir  }
     { and new Segment name.                                   }
     {=========================================================}
     FSegPath := AppendSlash(FSegDir) + ExtractFileName(FSegPath);

     {=========================================================}
     { If the user supplied an OnSegChange envent handler,     }
     { call it now and then check to see if they aborted the   }
     { operation                                               }
     {=========================================================}
     if (Assigned(FOnSegChange)) then
       OnSegChange(FSegPath);
     if (AbortOperation = True) then
       begin
       raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );
       exit;
       end;
end;

{=Create=======================================================}
{==============================================================}
constructor TSegLHA.Create(AOwner: TComponent);
var
  MsgCnt:    TSegMessageType;
begin
     inherited Create(AOwner);
     FSegBaseName := 'SegData';
     FSegDir := 'a:\';
     FSegExt := '';   {Change 2.50b 1/22/98}
     FSegSize :=1440000;
     AllowDuplicates := False;
     AbortOperation := False;
     FSegMessages := TStringList.Create;
     with FSegMessages do
       for MsgCnt := Low(TSegMessageType) to high(TSegMessageType) do
         add(SegMessageDefaults[MsgCnt]);
     FSegName := '';
     FSegNumber := 0;
     FSegPath := '';
     FTmpFile := '';
     FScanFile := '';
     FArcFile := '';
     FSegBytes := 0;
     FFreeBytes := 0;
     FTotalBytes := 0;
     FBytesProcessed := 0;
     FMyOwner := AOwner;                       {Bug Fix 9/28/97}
     GetMem(FBPtr,BUFSIZE);                    {Bug Fix 9/28/97}
end;

{=Destroy======================================================}
{==============================================================}
destructor TSegLHA.Destroy;
begin
{$IFDEF WIN32}
     FreeMem(FBPtr);                           {Bug Fix 9/28/97}
{$ELSE}
     FreeMem(FBPtr,BUFSize);                   {Bug Fix 9/28/97}
{$ENDIF}
     FSegMessages.Free;
     inherited Destroy;
end;

{=Compress=====================================================}
{==============================================================}
procedure TSegLHA.Compress;
begin
     AbortOperation := False;
     FSegNumber := 1;
     FSegBytes := 0;
     FFreeBytes := 0;
     FTotalBytes := 0;
     FBytesProcessed := 0;
     FArcFile := MakeTmpFile( '.$a$' );
     ArchiveName := FArcFile;
     try
       inherited Compress;
       CreateScanFile;
       SegmentArchive;
     except
       ArchiveName := '';
       RemoveSegFiles;
       raise;
     end; {try}
     ArchiveName := '';
     RemoveSegFiles;
end;

{=Expand=======================================================}
{==============================================================}
procedure TSegLHA.Expand;
begin
     AbortOperation := False;
     FSegNumber := 1;
     FSegBytes := 0;
     FFreeBytes := 0;
     FTotalBytes := 0;
     FBytesProcessed := 0;
     FArcFile := '';
     ArchiveName := FArcFile;
     try
       AssembleArchive(False);
     except
       RemoveSegFiles;
       raise;
     end; {try}
     ArchiveName := '';
     RemoveSegFiles;
end;

{=Scan=========================================================}
{==============================================================}
function  TSegLHA.Scan: Boolean;
begin
     Result := False;
     AbortOperation := False;
     FSegNumber := 1;
     FSegBytes := 0;
     FFreeBytes := 0;
     FTotalBytes := 0;
     FBytesProcessed := 0;
     FArcFile := '';
     ArchiveName := FArcFile;
     try
       ScanSegment;
     except
       RemoveSegFiles;
       raise;
     end; {try}
     ArchiveName := '';
     RemoveSegFiles;
     if (FileList.Count > 0) then
       Result := True;
end;

{=Verify=======================================================}
{==============================================================}
procedure TSegLHA.Verify;
begin
     AbortOperation := False;
     FSegNumber := 1;
     FSegBytes := 0;
     FFreeBytes := 0;
     FTotalBytes := 0;
     FBytesProcessed := 0;
     FArcFile := '';
     ArchiveName := FArcFile;
     try
       AssembleArchive(True);
     except
       RemoveSegFiles;
       raise;
     end; {try}
     ArchiveName := '';
     RemoveSegFiles;
end;

{=SegmentArchive===============================================}
{ This method segments the temporary archive created during    }
{ the Compress method. It assumes that the properties are all  }
{ in tact from the Compress operation.                         }
{==============================================================}
procedure TSegLHA.SegmentArchive;
var
   SetHdr:          TSegSetHeader;
   fh:              TFileStream;
begin
     {=========================================================}
     { Local initialization                                    }
     {=========================================================}
     SetHdr.HdrID := SSET;
     SetHdr.CompMode := CompressionMethod;
     SetHdr.SegDate := Now;
     SetHdr.SegSize := FSegSize;

     FTotalBytes := FTotalBytes + GetFileSize(FArcFile);

     {=========================================================}
     { Create the first segment file, write the TSegSetHeader; }
     { the Scan file and the Archive file to the segment stream}
     {=========================================================}
     fh := nil;
     try
       fh := CreateSegment;
       WriteToSegment(fh,SetHdr,SizeOf(TSegSetHeader));
       WriteScanFile(fh);
       WriteArchiveFile(fh);
     except
       if (fh <> nil) then
         fh.Free;
       raise;  {re-raise exception}
     end; {try}

     CloseSegment(fh);
end;

{=AssembleArchive==============================================}
{==============================================================}
procedure TSegLHA.AssembleArchive( VerifyOnly: Boolean );
var
  fh:               TFileStream;
  SetHdr:           TSegSetHeader;
  FL:               TStrings;
  NewTargetPath:    String;
begin
     {=========================================================}
     { Open the segment, read the SegSetHeader and the scan    }
     { file information                                        )
     {=========================================================}
     fh := nil;
     try
       fh := OpenSegment;
       ReadSegSetHeader(fh,SetHdr);
       ReadScanFile(fh);
     except
       if (fh <> nil) then
         CloseSegment(fh);
       raise;  {re-raise exception}
       exit;
     end; {try}

     {=========================================================}
     { Copy the file names from FileList into the property     }
     { FilesToProcess.  Then clear the FileList and the        }
     { FileInformation properties.                             }
     {=========================================================}
     FilesToProcess.Clear;
     FilesToProcess.Assign(FileList);
     FreeFileInfo; {from CompLHA}
     FileInformation.Clear;
     FileList.Clear;

     {=========================================================}
     { If the user supplied an OnSegExpand event handler, call }
     { it now and then apply and file name changes             }
     {=========================================================}
     if (Assigned(FOnSegExpand)) then
       begin
       FL := TStringList.Create;
       FL.AddStrings(FilesToProcess);
       NewTargetPath := TargetPath;
       OnSegExpand(TargetPath,NewTargetPath,FL);
       TargetPath := NewTargetPath;
       FilesToProcess.Clear;
       FilesToProcess.AddStrings(FL);
       FL.Clear;
       FL.Free;
       if (AbortOperation = True) then
         begin
         if (fh <> nil) then
           CloseSegment(fh);
         raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );
         exit;
         end;
       end;

     {=========================================================}
     { Assemble the archive by reading all the segments and    }
     { then call inherited Verify/Expand using the adjusted    }
     { FilesToPRocess and TargetPath, if the user changed them }
     {=========================================================}
     try
       ReadArchiveFile(fh);
       if (VerifyOnly) then
         inherited Verify
       else
         inherited Expand;
     except
       if (fh <> nil) then
         CloseSegment(fh);
       raise;  {re-raise exception}
       exit;
     end; {try}

     {=========================================================}
     { The inherited Expand does not raise an exception when   }
     { the operation is aborted. However the AbortOperation    }
     { flag is set. It is checked and an exception is raised   }
     { if it is True.                                          }
     {=========================================================}
     if (AbortOperation = True) then
       raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );

     CloseSegment(fh);
end;

{=ScanSegment==================================================}
{==============================================================}
procedure TSegLHA.ScanSegment;
var
  fh:         TFileStream;
  SetHdr:     TSegSetHeader;
begin
     {=========================================================}
     {=========================================================}
     fh := nil;
     try
       fh := OpenSegment;
       ReadSegSetHeader(fh,SetHdr);
       ReadScanFile(fh);
     except
       if (fh <> nil) then
         CloseSegment(fh);
       raise;  {re-raise exception}
     end; {try}

     CloseSegment(fh);
end;

{=ReadSegSetHeader===============================================}
{==============================================================}
procedure TSegLHA.ReadSegSetHeader( var fh: TFileStream;
                                        var SetHdr: TSegSetHeader );
begin
     FillChar(SetHdr,SizeOf(TSegSetHeader),$0);
     try
        ReadFromSegment(fh,SetHdr,SizeOf(TSegSetHeader));
     except
        raise; {re-raise exception}
     end;

     if (SetHdr.HdrID <> SSET) then
       raise ESegLHAError.Create( GetSegMessage(smSegSetHeaderMsg) );
end;

{=CreateScanFile===============================================}
{==============================================================}
procedure TSegLHA.CreateScanFile;
var
   fh:              TFileStream;
   i:               Integer;
   len:             Integer;
   BP:              PChar;
   FInfo:           TCompLHAFileInfo;
   FInfoRec:        TFileInfo;
   SaveFiles:       TStrings;
   SaveProcessed:   TStrings;
   SaveTargetPath:  String;
begin
     {=========================================================}
     { Local initialization                                    }
     {=========================================================}
     SaveFiles := TStringList.Create;
     SaveProcessed := TStringList.Create;
     SaveFiles.AddStrings(FilesToProcess);
     SaveProcessed.AddStrings(FilesProcessed);
     FilesToProcess.Clear;
     FilesProcessed.Clear;

     {=========================================================}
     { Call the inherited Scan method to populate the FileList }
     { and FileInformation properties. Then create a temp file }
     { that contains this data.                                }
     {=========================================================}
     try
       inherited Scan;
     except
       FilesToProcess.Clear;
       FilesProcessed.Clear;
       FilesToProcess.AddStrings(SaveFiles);
       FilesProcessed.AddStrings(SaveProcessed);
       raise;
       exit;
     end; {try}

     {=========================================================}
     { Using the FileList and FileInformation data, create a   }
     { temporary file.                                         }
     {=========================================================}
     FTmpFile := MakeTmpFile('.$t$');
     DeleteFile(FTmpFile);
     try
       fh := TFileStream.Create(FTmpFile,(fmCreate or fmOpenWrite));
     except
       raise;
     end; {try}

     {=========================================================}
     { Loop thru the FileList and FileInformation list and     }
     { write the data to the temporary file                    }
     {=========================================================}
     BP := FBPtr;                              {Bug Fix 9/28/97}
     for i := 0 to Pred(FileList.Count) do
       begin
       FillChar(BP^,BUFSIZE,$0);                       {Fill with nulls}
       BP^ := Char(Length(FileList[i]));               {Store name length}
       len := 1;
       StrPCopy((BP+len),FileList[i]);                 {Store name}
       len := len + Length(FileList[i]) + 1;           {Include a null}
       fh.WriteBuffer(BP^,len);                        {Write to temp file}
       FInfo := TCompLHAFileInfo(FileInformation[i]);  {Get File Info Object}

     {=========================================================}
     { Copy the FInfo data into a TFileInfo record. This is    }
     { a kludge to be able to write the object data to a file. }
     { **NOTE** If the TCompLHAFileInfo object changes, this   }
     { code will need to change as well.                       }
     {=========================================================}
       FInfoRec.CompressedMode := FInfo.CompressedMode;
       FInfoRec.CompressedSize := FInfo.CompressedSize;
       FInfoRec.FullSize := FInfo.FullSize;
       FInfoRec.Datetime := FInfo.Datetime;
       FInfoRec.Attributes := FInfo.Attributes;
       FInfoRec.HeaderLevel := FInfo.HeaderLevel;
       FInfoRec.CRC := FInfo.CRC;
       FInfoRec.Position := FInfo.Position;
       fh.WriteBuffer(FInfoRec,SizeOf(TFileInfo));
       end;

     fh.Free;

     {=========================================================}
     { Clear the FileList and FileInformation properties       }
     {=========================================================}
     FreeFileInfo; {from CompLHA}
     FileInformation.Clear;
     FileList.Clear;
     FilesToProcess.Clear;
     FilesProcessed.Clear;

     {=========================================================}
     { Compress the temporary file information file using the  }
     { same compression method the user specified.             }
     {=========================================================}
     DisableEventHandlers;
     FScanFile := MakeTmpFile('.$s$');
     if (FileExists(FScanFile) = True) then
       DeleteFile(FScanFile);
     ArchiveName := FScanFile;
     FilesToProcess.Add(FTmpFile);
     SaveTargetPath := TargetPath;
     TargetPath := TempArchivePath;
     try
       Inherited Compress;
     except
       TargetPath := SaveTargetPath;
       FilesToProcess.Clear;
       FilesToProcess.AddStrings(SaveFiles);
       FilesProcessed.Clear;
       FilesProcessed.AddStrings(SaveProcessed);
       SaveFiles.Free;
       SaveProcessed.Free;
       EnableEventHandlers;
       raise;
       exit;
     end; {try}

     {=========================================================}
     { Restore the properties and do any clean up              }
     {=========================================================}
     TargetPath := SaveTargetPath;
     FilesToProcess.Clear;
     FilesToProcess.AddStrings(SaveFiles);
     FilesProcessed.Clear;
     FilesProcessed.AddStrings(SaveProcessed);
     SaveFiles.Free;
     EnableEventHandlers;
end;

{=ReadScanFile================================================}
{==============================================================}
procedure TSegLHA.ReadScanFile( var fh: TFileStream );
var
  i:                Integer;
  len:              Integer;
  FInfo:            TCompLHAFileInfo;
  FInfoRec:         TFileInfo;
  ScanHdr:          TSegScanHeader;
  sfh:              TFileStream;
  BytesToRead:      Longint;
  BytesRemain:      Longint;
  BP:               PChar;
  SaveTargetPath:   String;
begin
     {=========================================================}
     { Read the SegScanHeader record and verify it             }
     {=========================================================}
     FillChar(ScanHdr,SizeOf(TSegScanHeader),$0);
     try
       ReadFromSegment( fh, ScanHdr, SizeOf(TSegScanHeader) );
     except
       raise; {re-raise exception}
       exit;
     end; {try}

     if (ScanHdr.HdrID <> SSCN) then
       begin
       raise ESegLHAError.Create( GetSegMessage(smSegScanHeaderMsg) );
       exit;
       end;

     {=========================================================}
     { Open the scan file                                      }
     {=========================================================}
     FArcFile := MakeTmpFile('.$a$');
     if (FileExists(FArcFile) = True) then
       DeleteFile(FArcFile);
     try
       sfh := TFileStream.Create( FArcFile, fmCreate or fmOpenWrite );
     except
       raise ESegLHAError.Create( GetSegMessage(smScanOpenMsg) );
       exit;
     end; {try}

     {=========================================================}
     { Get the compressed scan data from the segment stream    }
     { and write it to the temporary file                      }
     {=========================================================}
     BytesRemain := ScanHdr.ScanLen;
     BP := FBPtr;                              {Bug Fix 9/28/97}
     while (BytesRemain > 0) do
       begin
       FillChar(BP^,BUFSIZE,$0);
       BytesToRead := BUFSIZE;
       if (BytesToRead > BytesRemain) then
         BytesToRead := BytesRemain;
       try
         ReadFromSegment(fh,BP^,BytesToRead);
         sfh.Write(BP^,BytesToRead);
       except
         sfh.Free;
         raise; {re-raise exception}
         exit;
       end; {try}

       Dec(BytesRemain,BytesToRead);
       end; { while (BytesRemain > 0) }

     sfh.Free;

     {=========================================================}
     { Expand the temporary compressed scan file               }
     {=========================================================}
     DisableEventHandlers;
     FilesToProcess.Clear;
     ArchiveName := FArcFile;
     SaveTargetPath := TargetPath;
     TargetPath := TempArchivePath;;
     try
       inherited Expand;
     finally
       TargetPath := SaveTargetPath;
       FTmpFile := AnsiLowerCase(FilesProcessed[0]);
       FilesToProcess.Clear;
       RemoveFile(FArcFile);
       EnableEventHandlers;
     end; {try}

     {=========================================================}
     { Open the temporary scan list file                       }
     {=========================================================}
     try
       sfh := TFileStream.Create( FTmpFile, fmOpenRead );
     except
       raise ESegLHAError.Create( GetSegMessage(smScanOpenMsg) );
       exit;
     end; {try}

     {=========================================================}
     { Read the scan list file and populate the FileList and   }
     { FileInformation properties                              }
     {=========================================================}
     FileList.Clear;
     for i := 0 to pred(FileInformation.Count) do
       TCompLHAFileInfo(FileInformation[i]).free;
     FileInformation.Clear;

     BP := FBPtr;                              {Bug Fix 9/28/97}
     BytesRemain := sfh.Size;
     while(BytesRemain > 0) do
       begin
       FillChar(FBptr^,BUFSIZE,$0);            {Bug Fix 9/28/97}
       try
         sfh.Read(BP^,1);                  {read the length}
         len := Integer(BP^);              {convert the len}
         Dec(BytesRemain);
         sfh.Read((BP+1)^,len+1);          {read the filename + nul}
         Dec(BytesRemain,len+1);
         FileList.Add(StrPas((BP+1)));
         sfh.Read(FInfoRec,SizeOf(TFileInfo));
         Dec(BytesRemain,SizeOf(TFileInfo));
         FInfo := TCompLHAFileInfo.Create;
         FInfo.CompressedMode := FInfoRec.CompressedMode;
         FInfo.CompressedSize := FInfoRec.CompressedSize;
         FInfo.FullSize := FInfoRec.FullSize;
         FInfo.Datetime := FInfoRec.Datetime;
         FInfo.Attributes := FInfoRec.Attributes;
         FInfo.HeaderLevel := FInfoRec.HeaderLevel;
         FInfo.CRC := FInfoRec.CRC;
         FInfo.Position := FInfoRec.Position;
         FileInformation.Add(FInfo);

       except
         sfh.Free;
         raise;   {re-raise exception}
         exit;
       end; {try}
       end;

     sfh.Free;
end;

{=WriteScanFile================================================}
{==============================================================}
procedure TSegLHA.WriteScanFile( var fh: TFileStream );
var
  ScanHdr:       TSegScanHeader;
  sfh:           TFileStream;
  BytesRead:     Longint;
  BytesToRead:   Longint;
  BytesRemain:   Longint;
  BP:            PChar;
begin
     {=========================================================}
     { Open the scan file                                      }
     {=========================================================}
     try
       sfh := TFileStream.Create( FScanFile, fmOpenRead );
     except
       raise ESegLHAError.Create( GetSegMessage(smScanOpenMsg) );
       exit;
     end; {try}

     {=========================================================}
     { Write the TSegScanHeader to the segment stream          }
     {=========================================================}
     ScanHdr.HdrID := SSCN;
     ScanHdr.ScanLen := sfh.Size;
     try
       WriteToSegment( fh, ScanHdr, SizeOf(TSegScanHeader) );
     except
       sfh.Free;
      raise; {re-raise exception}
       exit;
     end; {try}

     {=========================================================}
     { Read the Scan file and write it to the Segment stream   }
     {=========================================================}
     BytesRemain := sfh.Size;
     BP := FBPtr;                              {Bug Fix 9/28/97}
     while (BytesRemain > 0) do
       begin
       FillChar(BP^,BUFSIZE,$0);
       BytesToRead := BUFSIZE;
       if (BytesToRead > BytesRemain) then
         BytesToRead := BytesRemain;
       try
         BytesRead := sfh.Read(BP^,BytesToRead);
         WriteToSegment(fh,BP^,BytesRead);
       except
         sfh.Free;
         raise; {re-raise exception}
         exit;
       end; {try}

       Dec(BytesRemain,BytesRead);
       end; { while (BytesRemain > 0) }

     sfh.Free;
end;

{=ReadArchiveFile==============================================}
{==============================================================}
procedure TSegLHA.ReadArchiveFile( var fh: TFileStream );
var
  ArcHdr:        TSegArcHeader;
  sfh:           TFileStream;
  BytesToRead:   Longint;
  BytesRemain:   Longint;
  BP:            PChar;
begin
     {=========================================================}
     { Create and Open the Archive file                        }
     {=========================================================}
     FArcFile := MakeTmpFile('.$a$');
     if (FileExists(FArcFile) = True) then
       DeleteFile(FArcFile);
     try
       sfh := TFileStream.Create( FArcFile, fmCreate or fmOpenWrite );
     except
       raise ESegLHAError.Create( GetSegMessage(smArcOpenMsg) );
       exit;
     end; {try}

     {=========================================================}
     { Read the TSegArcHeader from the segment stream          }
     {=========================================================}
     try
       ReadFromSegment( fh, ArcHdr, SizeOf(TSegArcHeader) );
     except
       sfh.Free;
       raise; {re-raise exception}
       exit;
     end; {try}

     {=========================================================}
     { Read from the segment stream and write to the archive   }
     {=========================================================}
     FTotalBytes := ArcHdr.ArcLen;
     FBytesProcessed := 0;
     BytesRemain := ArcHdr.ArcLen;
     BP := FBPtr;                              {Bug Fix 9/28/97}
     while (BytesRemain > 0) do
       begin
       FillChar(BP^,BUFSIZE,$0);
       BytesToRead := BUFSIZE;
       if (BytesToRead > BytesRemain) then
         BytesToRead := BytesRemain;
       try
         ReadFromSegment(fh,BP^,BytesToRead);
         sfh.Write(BP^,BytesToRead);
       except
         sfh.Free;
         raise; {re-raise exception}
         exit;
       end; {try}

       Dec(BytesRemain,BytesToRead);

     {=========================================================}
     { Update the FShowProgress and FSegShowProgress           }
     {=========================================================}
       Inc(FBytesProcessed,BytesToRead);
       FSaveOnShowProgress := OnShowProgress;
       if assigned(FSaveOnShowProgress) then
         begin
         if (FBytesProcessed > 100000) then
           FSegShowProgress := Integer(FBytesProcessed div (FTotalBytes div 100))
         else
           FSegShowProgress := Integer((100*FBytesProcessed) div FTotalBytes);
         OnShowProgress(FSegShowProgress);
         end
       else
         FSegShowProgress := 0;
       end; { while (BytesRemain > 0) }

     sfh.Free;
end;

{=WriteArchiveFile=============================================}
{==============================================================}
procedure TSegLHA.WriteArchiveFile( var fh: TFileStream );
var
  ArcHdr:        TSegArcHeader;
  sfh:           TFileStream;
  BytesRead:     Longint;
  BytesToRead:   Longint;
  BytesRemain:   Longint;
  BP:            PChar;
begin
     {=========================================================}
     { Open the Archive file                                   }
     {=========================================================}
     try
       sfh := TFileStream.Create( FArcFile, fmOpenRead );
     except
       raise ESegLHAError.Create( GetSegMessage(smArcOpenMsg) );
       exit;
     end; {try}

     {=========================================================}
     { Write the TSegArcHeader to the segment stream           }
     {=========================================================}
     ArcHdr.HdrID := SARC;
     ArcHdr.ArcLen := sfh.Size;
     try
       WriteToSegment( fh, ArcHdr, SizeOf(TSegArcHeader) );
     except
       sfh.Free;
       raise; {re-raise exception}
       exit;
     end; {try}

     {=========================================================}
     { Read the Archive file and write it to the Segment stream}
     {=========================================================}
     FTotalBytes := sfh.Size;
     FBytesProcessed := 0;
     BytesRemain := sfh.Size;
     BP := FBPtr;                              {Bug Fix 9/28/97}
     while (BytesRemain > 0) do
       begin
       FillChar(BP^,BUFSIZE,$0);
       BytesToRead := BUFSIZE;
       if (BytesToRead > BytesRemain) then
         BytesToRead := BytesRemain;
       try
         BytesRead := sfh.Read(BP^,BytesToRead);
         WriteToSegment(fh,BP^,BytesRead);
       except
         sfh.Free;
         raise; {re-raise exception}
         exit;
       end; {try}

       Dec(BytesRemain,BytesRead);

     {=========================================================}
     { Update the FShowProgress and FSegShowProgress           }
     {=========================================================}
       Inc(FBytesProcessed,BytesRead);
       FSaveOnShowProgress := OnShowProgress;
       if assigned(FSaveOnShowProgress) then
         begin
         if (FBytesProcessed > 100000) then
           FSegShowProgress := Integer(FBytesProcessed div (FTotalBytes div 100))
         else
           FSegShowProgress := Integer((100*FBytesProcessed) div FTotalBytes);
         OnShowProgress(FSegShowProgress);
         end
       else
         FSegShowProgress := 0;

       end; { while (BytesRemain > 0) }

     sfh.Free;
end;

{=CreateSegment================================================}
{==============================================================}
function  TSegLHA.CreateSegment: TFileStream;
var
  FreeBytes:        Single; { was Longint v2.50b }
  fh:               TFileStream;
  SegIDHdr:         TSegIDHeader;
  Msg:              String;
begin
     {=========================================================}
     { Prompt the user for the next segments media.  If there  }
     { is not enough free space on the media, we stay in this  }
     { loop until media with enough space is used or until the }
     { users aborts the operation.                             }
     {=========================================================}
     fh := nil;
     SetSegmentPath;
     while(True) do
       begin

       {=========================================================}
       { The NewSegmentEvent method handles the segment prompts  }
       {=========================================================}
       try
         NewSegmentEvent;
       except
         raise;  {re-raise exception}
         exit;
       end; {try}

       {=========================================================}
       { FSegPath now contains the drive and path to the new     }
       { segment.  Delete the segment file if it exists and then }
       { check the free space for the minimum required.          }
       {=========================================================}
       if (FileExists(FSegPath) = True) then
         DeleteFile(FSegPath);
       FreeBytes := CheckSpace( UpperCase(FSegPath) );
       if (FreeBytes > MINFREE) then
         begin
         FSegBytes := FSegSize;
         if (FreeBytes < FSegSize) then
           FSegBytes := Trunc(FreeBytes); { v2.50b }
         break; { break out of while loop }
         end;

       Msg := GetSegMessage(smFreeSpaceMsg) + CR + GetSegMessage(smCorrectCancelMsg);
       if (MessageDlg(Msg,mtWarning,[mbCancel,mbOK],0) = mrCancel) then
         begin
         AbortOperation := True;
         raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );
         exit;
         end;

       end; { while(True) do }

     {===========================================================}
     { Try to Create/Open the requested segment                  }
     {===========================================================}
     try
       ForceDirectories(FSegDir);
       fh := TFileStream.Create( FSegPath,fmCreate or fmOpenWrite);
       SegIDHdr.HdrID := SSID;
       SegIDHdr.SegNumber := FSegNumber;
       fh.WriteBuffer(SegIDHdr,SizeOf(TSegIDHeader));
       Dec(FSegBytes,SizeOf(TSegIDHeader));
     except
       if (fh <> nil) then
         begin
         fh.Free;
         end;
       raise ESegLHAError.Create( GetSegMessage(smSegOpenMsg) );
       exit;
     end; {try}
     Result := fh;
end;

{=OpenSegment==================================================}
{==============================================================}
function  TSegLHA.OpenSegment: TFileStream;
var
  fh:               TFileStream;
  IDHdr:            TSegIDHeader;
  BP:               PChar;
  Msg:              String;
begin
     {=========================================================}
     { Prompt the user for the next segments media.  We stay   }
     { in loop until the user puts in the correct media or the }
     { user aborts the operation.                              }
     {=========================================================}
     SetSegmentPath;
     while(True) do
       begin

       {=========================================================}
       { The NewSegmentEvent method handles the segment propmts  }
       {=========================================================}
       try
         NewSegmentEvent;
       except
         raise;  {re-raise exception}
         exit;
       end; {try}

       {=========================================================}
       { New in 2.50b 1/22/98.  If the segment exists, open it   }
       { and read the SegIDHeader to check if it is the correct  }
       { segment.                                                }
       {=========================================================}
       if (FileExists(FSegPath) = True) then
         begin
         try
           fh := TFileStream.Create( FSegPath,fmOpenRead );
           FillChar(IDHdr,SizeOf(TSegIDHeader),$0);
           BP := PChar(@IDHdr);
           fh.ReadBuffer(BP^,SizeOf(TSegIDHeader));
           fh.Free;
           if ((IDHdr.HdrID = SSID) and (IDHdr.SegNumber = FSegNumber)) then
             break  { Break out of the loop}
           else
             Msg := GetSegMessage(smWrongSegMounted) + CR +
                                   GetSegMessage(smCorrectCancelMsg);
         except
           fh := nil;
           raise ESegLHAError.Create( GetSegMessage(smSegOpenMsg) );
           exit;
         end; {try}
         end
       else
         Msg := GetSegMessage(smSegNotFoundMsg) + CR + GetSegMessage(smCorrectCancelMsg);
       {end of 2.50b 1/22/98 change}

       if (MessageDlg(Msg,mtWarning,[mbCancel,mbOK],0) = mrCancel) then
         begin
         AbortOperation := True;
         raise ECompLHAOperationAborted.Create( GetSegMessage(smSegAbortMsg) );
         exit;
         end;

       end; { while(True) do }

     {===========================================================}
     { Try to Create/Open the requested segment                  }
     {===========================================================}
     try
       fh := TFileStream.Create( FSegPath,fmOpenRead );
       FSegBytes := fh.Size;
     except
       raise ESegLHAError.Create( GetSegMessage(smSegOpenMsg) );
       exit;
     end; {try}

     {===========================================================}
     { Read the Segment ID record and verify we have the right   }
     { one                                                       }
     {===========================================================}
     try
       FillChar(IDHdr,SizeOf(TSegIDHeader),$0);
       ReadFromSegment(fh,IDHdr,SizeOf(TSegIDHeader));
     except
       fh.Free;
       fh := nil;
       raise;
       exit;
     end; {try}

     if (IDHdr.HdrID <> SSID) then
       begin
       fh.Free;
       fh := nil;
       raise ESegLHAError.Create( GetSegMessage(smSegIDHeaderMsg) );
       end;

     Result := fh;
end;

{=CloseSegment=================================================}
{==============================================================}
procedure TSegLHA.CloseSegment( var fh: TFileStream );
begin
     if (fh <> nil) then
       fh.Free;
     fh := nil;
end;

{=WriteToSegment===============================================}
{==============================================================}
procedure TSegLHA.WriteToSegment( var fh: TFileStream; var Buffer;
                                     Count: Longint );
var
   BytesRemain:       Longint;
   BytesToWrite:      Longint;
   pos:               Longint;
   BP:                PChar;
begin
     BP := PChar(@Buffer);
     pos := 0;
     BytesRemain := Count;

     {=========================================================}
     { Main WriteToSegment loop. Stay in this loop until the   }
     { BytesRemain = 0.                                        }
     {=========================================================}
     while (BytesRemain > 0) do
       begin

       {=======================================================}
       { If FSegBytes = 0, then we need to create the next     }
       { segment.                                              }
       {=======================================================}
       if (FSegBytes = 0) then
         begin
         fh.Free;    {Close the current segment}
         fh := nil;

         {=====================================================}
         { Create and open the next segment                    }
         {=====================================================}
         Inc(FSegNumber);
         try
           fh := CreateSegment;
         except
           raise;  {re-raise the exception}
           exit;
         end; {try}
         end; { if (FSegBytes = 0) }

       {=======================================================}
       { Write the buffer to the segment stream                }
       {=======================================================}
       BytesToWrite := BUFSIZE;
       if (BytesRemain < BytesToWrite) then
         BytesToWrite := BytesRemain;
       if (BytesToWrite > FSegBytes) then
         BytesToWrite := FSegBytes;
       try
         fh.WriteBuffer((BP+pos)^,BytesToWrite);
       except
         raise; { re-raise exception }
         exit;
       end; {try}

       Inc(pos,BytesToWrite);
       Dec(FSegBytes,BytesToWrite);
       Dec(BytesRemain,BytesToWrite);
       {User should call Application.ProcessMessages}
       end; { while (BytesRemain > 0) do }
end;

{=ReadFromSegment==============================================}
{==============================================================}
procedure TSegLHA.ReadFromSegment( var fh: TFileStream; var Buffer;
                                       Count: Longint );
var
   BytesRemain:       Longint;
   BytesToRead:       Longint;
   pos:               Longint;
   BP:                PChar;
begin
     BP := PChar(@Buffer);
     pos := 0;
     BytesRemain := Count;

     {=========================================================}
     { Main ReadFromSegment loop. Stay in this loop until the  }
     { BytesRemain = 0.                                        }
     {=========================================================}
     while (BytesRemain > 0) do
       begin

       {=======================================================}
       { If FSegBytes = 0, then we need to open the next       }
       { segment.                                              }
       {=======================================================}
       if (FSegBytes = 0) then
         begin
         fh.Free;    {Close the current segment}
         fh := nil;

         {=====================================================}
         { Open the next segment                               }
         {=====================================================}
         Inc(FSegNumber);
         try
           fh := OpenSegment;
         except
           raise;  {re-raise the exception}
           exit;
         end; {try}
         end; { if (FSegBytes = 0) }

       {=======================================================}
       { Read the request number of bytes from the segment     }
       { stream into the buffer                                }
       {=======================================================}
       BytesToRead := BUFSIZE;
       if (BytesRemain < BytesToRead) then
         BytesToRead := BytesRemain;
       if (BytesToRead > FSegBytes) then
         BytesToRead := FSegBytes;
       try
         fh.ReadBuffer((BP+pos)^,BytesToRead);
       except
         raise; { re-raise exception }
         exit;
       end; {try}

       Inc(pos,BytesToRead);
       Dec(FSegBytes,BytesToRead);
       Dec(BytesRemain,BytesToRead);
       {User should call Application.ProcessMessages}
       end; { while (BytesRemain > 0) do }
end;

end.
